/**
 * file IEL_exec_info.h
 */

#ifndef _IEL_EXEC_INFO_H
#define _IEL_EXEC_INFO_H

#if defined(DEBUG) && DEBUG > 0
 #define DBG(fmt, args...) fprintf(stdout, "IEL_DEBUG: %s:%d:%s(): " fmt "\n", \
         __FILE__, __LINE__, __func__, ##args); fflush(stdout)
#define LINE DBG()
#else
 #define DBG(fmt, args...) /* Don't do anything in release builds */
 #define LINE
#endif

/* Macros to access the current module's argc and argv from confic file */
#define IEL_ARGC(a) a->modules[a->module_num].mod_argc
#define IEL_ARGV(a) a->modules[a->module_num].mod_argv


#include "IEL.h"

#ifdef __cplusplus
extern "C" {
#endif

/**
 * This structure represents information required to describe workflow groups,
 * specified in the congiguration file.
 */

typedef struct {
  int size;
  int iterations;
  int persist;
  int num_modules;
  char **order;
} IEL_workflow_group_t;

/**
 * This type represents whether a module is serial or parallel
 */

typedef enum  exec_mode_t { SERIAL, PARALLEL } exec_mode_t;

/**
 * This structure represents the module information from the configuration
 * file.
 */
typedef struct {
  int***  points;  /**< Array of points that this module 'owns' */
  double* input;  /**< Initial values for data */
  char*   libname;  /**< Name of the shared library to load */
  char*   libtype;
  char    splitdir; //Boolean, do we need separate working directories?
  char*   splitdirpath; //Path to separate directories
  char*   funcname; /**< Function name of the module to be called */
  char*   stdinFile; /**< File to redirect stdin from */
  int     emode; /**< Is this module SERIAL or PARALLEL? (execution mode)*/
  int     mod_argc;		/**< The number of char * module arguments */
  char**  mod_argv;	/**< Array of char * module arguments */
} module_depend_t;

/*! \var typedef double shared_bc_t
    \brief Type definition for shared_bc_t

    The shared_bc_t type can be changed to represent the kind of data
    you expect to use.
*/
typedef double shared_bc_t;

/**
 * This structure represents everything the IEL Executive needs to know
 * to operate.
 */
typedef struct {
  int  num_shared_bc;     /**< Number of shared BCs */
  int* shared_bc_sizes;  /**< Array of sizes for each shared BC */
  int* bc_start;         /**< start idx of each shared BC */
  int* bc_len;           /**< Length (number of points) of each SBC */

  int  num_modules;    /**< Number of modules */
  int* module_sizes;  /**< Array of sizes of each module */
  int  num_work_groups;    /**< Number of workflow groups */

  module_depend_t* modules;  /**< Module info struct for each module */
  IEL_workflow_group_t * work_groups;

  int   packsize;  /**< size of packed buffer needed to xmit IEL exec struct */
  char* packed;  /**< packed buffer for xmitting IEL exec struct vi MPI */

  MPI_Comm IEL_comm;   /**< MPI communicator for the whole IEL system */
  MPI_Comm module_comm; /**< MPI communicator for just the modules */
  MPI_Comm group_comm; /**< MPI communicator for just the group */
  
  int IEL_rank;        /**< MPI rank within the IEL_comm communicator */
  int module_rank;      /**< MPI rank within the module_comm communicator */
  int IEL_num_ranks;   /**< Number of MPI ranks in IEL_comm communicator */
  int module_num_ranks; /**< Number of MPI ranks in module_comm communicator */
  int module_num;       /**< Current module number */
  int work_group_num;   /**< Current number of work group*/
  int num_proc;         /**< Total number of processes for all modules */

  shared_bc_t** shared_bc;           /**< The shared BC data */

  int num_handles;                   /**< Number of IELCOMM data handles */
  IEL_data_handle_t** handles;       /**< Array of IELCOMM data handles*/

  int*** commtab;                    /**< Table of procs with shared points [for internal use only] */
  
  IEL_component_info_t* cinfo;      /**< Pointer to the IELCOMM structure */

  int tuple_size;                   /**< Number of tuple space servers */

  shared_bc_t* accessible_bc;	    /**< Accessible elements of shared_bc; to lessen amount sent and received */
  int accessible_bc_size;	    /** Number of elements in accessible_bc */
  int uses_mvmt;		    /**< Determines if movement options are used */
  int*** sources;		    /**< Sources of each shared_bc*/
  int*** dests;			    /**< Destinations of each shared_bc*/

  //char* cfg_filename;             /** Name of the config file **/
} IEL_exec_info_t;


//Just the essential information from the struct above
//for a module using tuple space communication.
typedef struct {
	int      num_modules;
	int*     module_sizes;
	int      mod_argc;
	char**   mod_argv;
	MPI_Comm IEL_comm;
	MPI_Comm module_comm;
	int      IEL_rank;
	int      module_rank;
	int      IEL_num_ranks;
	int      module_num_ranks;
	int      module_num;
	int      num_proc;
	int      tuple_size;
} IEL_reduced_info_t;

//For a Fortran module we need to separately pass
//a converted argv array, so we don't include it here.
typedef struct {
	int      num_modules;
	MPI_Comm IEL_comm;
	MPI_Comm module_comm;
	int      IEL_rank;
	int      module_rank;
	int      IEL_num_ranks;
	int      module_num_ranks;
	int      module_num;
	int      num_proc;
	int      tuple_size;
} IEL_fort_info_t;


char* dupstr(const char* str);
int IELExecInit(int, char**);
int IELExecutive(MPI_Comm, char*);
void IELAddModule(int (*func)(IEL_exec_info_t*), char* func_name);
int IEL_exec_config(IEL_exec_info_t*, char*);
int IEL_exec_clear(IEL_exec_info_t*);
int IEL_exec_init_from_hdf5(IEL_exec_info_t*, char*);
int IEL_exec_dump_info(IEL_exec_info_t*);
int IEL_exec_unpack_info(IEL_exec_info_t*, MPI_Comm);
int IEL_exec_pack_info(IEL_exec_info_t*, MPI_Comm);
int IEL_exec_init_handles(IEL_exec_info_t*);
int IEL_exec_get_idx(IEL_exec_info_t*, int, int*, int*);
IEL_fort_info_t reduceFortInfo(IEL_exec_info_t*);
int IEL_workflow_init(IEL_exec_info_t *, int, char *);
int IEL_recv(IEL_exec_info_t*, int);
int IEL_send(IEL_exec_info_t*, int);
int IEL_move(IEL_exec_info_t*, int);
int IEL_barrier(IEL_exec_info_t*);
int IEL_module_barrier(IEL_exec_info_t*);
int IEL_group_barrier(IEL_exec_info_t*);
void IEL_insert_bc(IEL_exec_info_t*, shared_bc_t*, int, int, int, int);
void IEL_copy_bc(IEL_exec_info_t*, shared_bc_t*, int, int, int, int);

void IEL_exec_print_sbc(IEL_exec_info_t *);

#ifdef __cplusplus
}
#endif

#endif
